/*	
 *	clipmac.c
 *
 *	P/ECE USB Ethernet MACw
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2005 Naoyuki Sawa
 *
 *	* Thu Aug 18 20:36:00 JST 2005 Naoyuki Sawa
 *	- 1st [XB
 */
#include "clip.h"

/****************************************************************************
 *	fBXNv^`
 ****************************************************************************/

/* foCXfBXNv^ */
static const DEVICE_DESCRIPTOR device_desc = {
	bLength:		sizeof(DEVICE_DESCRIPTOR),
	bDescriptorType:	DEVICE_DESCRIPTOR_TYPE,
	bcdUSB:			0x0110,
	bMaxPacketSize0:	EP0_PACKET_SIZE,
	idVendor:		PCE_VENDOR_ID,
	idProduct:		MAC_PRODUCT_ID,				/* d֎~!! */
	bcdDevice:		0x0001,
	iManufacture:		1,
	iProduct:		2,
	bNumConfigurations:	1,
};

/* RtBM[VfBXNv^ */
static const CONFIGURATION_DESCRIPTOR config_desc = {
	bLength:		sizeof(CONFIGURATION_DESCRIPTOR),
	bDescriptorType:	CONFIGURATION_DESCRIPTOR_TYPE,
	wTotalLength:		sizeof(CONFIGURATION_DESCRIPTOR) +	/* config_desc */
				sizeof(INTERFACE_DESCRIPTOR    ) +	/* interface_desc */
				sizeof(ENDPOINT_DESCRIPTOR     ) +	/* endpoint_desc_ep2in */
				sizeof(ENDPOINT_DESCRIPTOR     ),	/* endpoint_desc_ep2out */
	bNumInterfaces:		1,
	bConfigurationValue:	1,
	iConfiguration:		3,
	bmAttributes:		0x80,					/* Bus-powered */
	MaxPower:		100 / 2,				/* 100[mA] () */
};

/* C^[tFCXfBXNv^ */
static const INTERFACE_DESCRIPTOR interface_desc = {
	bLength:		sizeof(INTERFACE_DESCRIPTOR),
	bDescriptorType:	INTERFACE_DESCRIPTOR_TYPE,
	bNumEndpoints:		2,
	bInterfaceClass:	0xff,					/* Vendor-specific */
	bInterfaceSubClass:	0xff,					/* Vendor-specific */
	bInterfaceProtocol:	0xff,					/* Vendor-specific */
	iInterface:		4,
};

/* Gh|CgfBXNv^
 * * P/ECERpC̕s΍̂߁Aendpoint_desc\̔zɂȂ悤!!
 *   ڂ́Aclip/clipusb.h̃RgQƂĂB
 */
static const ENDPOINT_DESCRIPTOR endpoint_desc_ep2in = {
	bLength:		sizeof(ENDPOINT_DESCRIPTOR),
	bDescriptorType:	ENDPOINT_DESCRIPTOR_TYPE,
	bEndpointAddress:	0x82,					/* EP2IN */
	bAttributes:		2,					/* Bulk */
	wMaxPacketSize:		EP2_NONISO_PACKET_SIZE,
};
static const ENDPOINT_DESCRIPTOR endpoint_desc_ep2out = {
	bLength:		sizeof(ENDPOINT_DESCRIPTOR),
	bDescriptorType:	ENDPOINT_DESCRIPTOR_TYPE,
	bEndpointAddress:	0x02,					/* EP2OUT */
	bAttributes:		2,					/* Bulk */
	wMaxPacketSize:		EP2_NONISO_PACKET_SIZE,
};

/* XgOfBXNv^
 * * gcćAL"..."`16bit UNICODEg܂B(32bitɂȂĂ܂܂)
 *   Ȃ̂ŁA1oCgÂLq܂B
 * * 񒆂[0-7]܂ނƁA'\0'̑ƌȂ̂ŁAӂĂB
 */
static const char* const string_desc[] = {
//	   ++--- 04 06 08 0a 0c 0e 10 12 14 16 18 1a 1c 1e 20 22 24 26 28 2a 2c 2e 30
//	   ||
//	"\x??\3_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0_\0"
	"\x04\3\x09\x04",								// 0: wLANGID = 0x0409 English (United States)
	"\x16\3P\0i\0e\0c\0e\0 \0L\0a\0b\0.\0",						// 1: iManufacture   = "Piece Lab."
	"\x22\3P\0/\0E\0C\0E\0 \0U\0S\0B\0 \0d\0e\0v\0i\0c\0e\0",			// 2: iProduct       = "P/ECE USB device"
	"\x2e\3E\0t\0h\0e\0r\0n\0e\0t\0 \0c\0o\0n\0f\0i\0g\0u\0r\0a\0t\0i\0o\0n\0",	// 3: iConfiguration = "Ethernet configuration"
	"\x26\3E\0t\0h\0e\0r\0n\0e\0t\0 \0i\0n\0t\0e\0r\0f\0a\0c\0e\0",			// 4: iInterface     = "Ethernet interface"
};

/****************************************************************************
 *	
 ****************************************************************************/

/* hCo\ */
typedef struct _MAC {
	unsigned char address[6];	// MACAhX (vendor_requestŎݒ)

	QUEUE* rxque;			// ML[
	unsigned char rxbuf[1514];	// MpPbgobt@
	int rxlen;			// MpPbgTCY
	int rxok;			// MpPbg
	int rxerr;			// MG[̃pPbg
	int rxovr;			// MI[o[̃pPbg
	void (*rxdone)();		// MR[obN (NULL:)

	QUEUE* txque;			// ML[
	unsigned char txbuf[1514];	// MpPbgobt@
	int txlen;			// MpPbgTCY
	int txpos;			// MpPbg|C^
	int txing;			// M~=0, 쒆=1
	int txok;			// MpPbg
	int txerr;			// MG[̃pPbg (2005/08/19݁Agp)
	int txovr;			// MI[o[̃pPbg
	void (*txdone)();		// MR[obN (NULL:)

	unsigned char cxbuf[sizeof config_desc	// foCXNGXgp̈ꎞobt@
			  + sizeof interface_desc
			  + sizeof endpoint_desc_ep2in
			  + sizeof endpoint_desc_ep2out];
} MAC;
static MAC mac;

static void get_descriptor(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void set_configuration(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void set_interface(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void vendor_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len);
static void ep2_rxdone(const unsigned char dat[/*len*/], int len);
static void ep2_txdone();
static int  ep2_txsub();

/* USBtbN */
static const USBINF usb_inf = {
	/* USBC mode */
	mode:			D12_ENDP_NONISO,
	/* USBC raw event */
	ep2_rxdone:		ep2_rxdone,
	ep2_txdone:		ep2_txdone,
	/* Device request */
	get_descriptor:		get_descriptor,
	set_configuration:	set_configuration,
	set_interface:		set_interface,
	vendor_request:		vendor_request,
};

/* x_NGXgR[h (foCXhCo̒`ƈv邱) */
#define SET_DEVICE_ADDRESS	0	/* MACAhXݒ肵܂ */

/* MACu[hLXgAhX(FF-FF-FF-FF-FF-FF) */
const unsigned char mac_broadcast_address[6] = { 0xff,0xff,0xff,0xff,0xff,0xff };

/****************************************************************************
 *	
 ****************************************************************************/

static void
get_descriptor(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	int type  = HIBYTE(req->wValue);
	int index = LOBYTE(req->wValue);

	/* vꂽfBXNv^ԑ܂B */
	switch(type) {
	case DEVICE_DESCRIPTOR_TYPE:
		usb_ep0_tx(&device_desc, sizeof device_desc);
		return;
	case CONFIGURATION_DESCRIPTOR_TYPE:
		switch(index) {
		case 0:
			len = 0;
			memcpy(&mac.cxbuf[len], &config_desc         , sizeof config_desc         ); len += sizeof config_desc         ;
			memcpy(&mac.cxbuf[len], &interface_desc      , sizeof interface_desc      ); len += sizeof interface_desc      ;
			memcpy(&mac.cxbuf[len], &endpoint_desc_ep2in , sizeof endpoint_desc_ep2in ); len += sizeof endpoint_desc_ep2in ;
			memcpy(&mac.cxbuf[len], &endpoint_desc_ep2out, sizeof endpoint_desc_ep2out); len += sizeof endpoint_desc_ep2out;
			usb_ep0_tx(mac.cxbuf, len);
			return;
		}
		break;
	case STRING_DESCRIPTOR_TYPE:
		if(index < ARRAY_SIZE(string_desc)) {
			usb_ep0_tx(string_desc[index], string_desc[index][0]);
			return;
		}
		break;
	}
	usb_ep0_stall(); /* Ή */
}

static void
set_configuration(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	/* RtBO[V#1(ꂵ܂)IꂽACGh|CgLɂ܂B */
	D12_SetEndpointEnable(req->wValue == 1);
	usb_ep0_tx(NULL, 0); /* ACK */
}

static void
set_interface(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	/* C^[tFCX̑֐ݒ͂ЂƂȂ̂ŁA܂B */
	usb_ep0_tx(NULL, 0); /* ACK */
}

static void
vendor_request(const DEVICE_REQUEST* req, const unsigned char* dat, int len)
{
	switch(req->bRequest) {
	case SET_DEVICE_ADDRESS:
		/* MACAhXݒ肵܂B */
		if(len == 6) {
			memcpy(mac.address, dat, 6);
			usb_ep0_tx(NULL, 0); /* ACK */
		} else {
			usb_ep0_stall();
		}
		return;
	}
	usb_ep0_stall();
}

/****************************************************************************
 *	
 ****************************************************************************/

static void
ep2_rxdone(const unsigned char dat[/*len*/], int len)
{
	int retval;

	/* Mf[^pPbgobt@֒ǉ܂B
	 * őt[TCY߂̏ꍇ́Af[^̂āATCŶݑ܂B
	 * pPbgMɁAvTCYApPbg𔻒f邽߂łB
	 */
	if(mac.rxlen + len <= 1514) {
		memcpy(mac.rxbuf + mac.rxlen, dat, len);
	}
	mac.rxlen += len;

	/* Gh|CgTCỸf[^MɂApPbgM܂B
	 * pPbgTCYGh|CgTCY̔{łꍇ́A
	 * Ōɋf[^ė͂Ȃ̂ŁAŊ𔻒fł܂B
	 */
	if(len < EP2_NONISO_PACKET_SIZE) {

		/* pPbǵAG[JEgAPɖ邱Ƃɂ܂B
		 * PChCoAGh|CgTCỸf[^M̌ɁA
		 * ɔOċpPbg𑗂ė\邩łB
		 */
		if(!mac.rxlen) {
			/** no job **/
			goto L_SKIP;
		}

		/* MpPbǵAt[TCY܂B */
		if((mac.rxlen < 60) || (mac.rxlen > 1514)) {
			mac.rxerr++; /* t[TCYs */
			goto L_SKIP;
		}

		/* {̃C[T[lbgȂ΁ÃzXgẴt[M邱ƂL蓾܂A
		 * P/ECE<->PC̍\̓sAtosAȂ̂ŁAẴt[ėȂ͂łB
		 * fobOeՂɂ邽߁AĈȊOŁAAu[hLXgȊÕt[A
		 * MG[ƂăJEg邱Ƃɂ܂B
		 */
		if((memcmp(mac.rxbuf, mac.address,           6) != 0) &&
		   (memcmp(mac.rxbuf, mac_broadcast_address, 6) != 0)) {
			//mac.rxerr++; /* ĐAhXs */
			//{{2005/08/18 Naoyuki Sawa}}
			//ςAG[JEgȂƂɂ܂B
			//΂炭ŗlqāAǂ邩߂܂B
			goto L_SKIP;
		}

		/* MpPbgAML[֒ǉ܂B */
		retval = append_queue(mac.rxque, mac.rxbuf, mac.rxlen);
		if(retval < 0) {
			mac.rxovr++; /* MI[o[ */
			goto L_SKIP;
		}

		/* MpPbgAJEgAbv܂B */
		mac.rxok++;

		/* KvɉāAMR[obN֒ʒm܂B */
		if(mac.rxdone) {
			mac.rxdone();
		}
L_SKIP:
		/* Mobt@ɖ߂܂B */
		mac.rxlen = 0;
	}
}

static void
ep2_txdone()
{
	/* MpPbgAGh|Cg֏݂܂B */
	ASSERT(mac.txing);
	mac.txing = ep2_txsub();

	/* M~ֈڍs... */
	if(!mac.txing) {

		/* MR[obN΁AĂяo܂B */
		if(mac.txdone) {
			mac.txdone();
			/* txdone()mac_send()Ă񂾏ꍇAmac_send()oRep2_txsub()ɂA
			 * VMpPbgGh|Cgɏ܂Atxing=1ɖ߂Ă܂B
			 */
		}
	}
}

/* M̃pPbgAGh|Cg֏݂܂B
 * M̃pPbg΁AL[玟̑MpPbg擾AGh|Cg֏݂܂B
 * L[Ȃ΁A܂B
 * Gh|Cgւ݂̏sꍇ́A      1(M쒆)Ԃ܂B
 * Gh|Cgւ݂̏sȂꍇ́A0(M~)Ԃ܂B
 */
static int
ep2_txsub()
{
	int len;

	/* M̃pPbg... */
	if(!mac.txlen) {

		/* ML[玟̑MpPbg擾܂B */
		len = remove_queue(mac.txque, mac.txbuf, sizeof mac.txbuf);
		if(len < 0) {

			/* ML[Ȃ΁AA܂B */
			return 0; /* M~ */
		}

		/* ŏt[TCỸpPbǵAŏt[TCY֊g܂B */
		if(len < 60) {
			len = 60;
		}

		/* MpPbgTCYƃ|C^܂B */
		mac.txlen = len;
		mac.txpos = 0;
	}

	/* MpPbgAGh|Cg֏݂܂B */
	len = mac.txlen - mac.txpos; /* len=0L蓾܂ */
	if(len > EP2_NONISO_PACKET_SIZE) {
		len = EP2_NONISO_PACKET_SIZE;
	}
	D12_WriteEndpoint(EP2IN, mac.txbuf + mac.txpos, len);
	mac.txpos += len;

	/* Gh|CgTCY̑MŁApPbgI[܂B */
	if(len < EP2_NONISO_PACKET_SIZE) {
		mac.txlen = 0;

		/* MpPbgAJEgAbv܂B */
		mac.txok++;
	}

	return 1; /* M쒆 */
}

/****************************************************************************
 *	
 ****************************************************************************/

#define MIN_QUEUE_CAPACITY (sizeof(int) + 1514) /* ML[̍ŏTCY */

void
mac_open(int rxque_capacity, int txque_capacity, void (*rxdone)(), void (*txdone)())
{
	/* ܂Amɐؒf܂B */
	mac_close();

	/* hCo\̂NA܂B */
	memset(&mac, 0, sizeof mac);

	/* ML[쐬AR[obNi[܂B */
	if(rxque_capacity < MIN_QUEUE_CAPACITY) {
		rxque_capacity = MIN_QUEUE_CAPACITY;
	}
	mac.rxque = create_queue(rxque_capacity);
	mac.rxdone = rxdone; 

	/* ML[쐬AR[obNi[܂B */
	if(txque_capacity < MIN_QUEUE_CAPACITY) {
		txque_capacity = MIN_QUEUE_CAPACITY;
	}
	mac.txque = create_queue(txque_capacity);
	mac.txdone = txdone;

	/* USBtbN܂B */
	usb_init(&usb_inf);
}

void
mac_close()
{
	/* USB̃tbN܂B */
	usb_free();

	/* ML[폜܂B */
	if(mac.rxque) {
		delete_queue(mac.rxque);
	}

	/* ML[폜܂B */
	if(mac.rxque) {
		delete_queue(mac.rxque);
	}

	/* hCo\̂NA܂B */
	memset(&mac, 0, sizeof mac);
}

int
mac_recv(void* packet/*[maxlength]*/, int maxlength)
{
	int retval;

	if(maxlength < 0) {
		DIE();
	}

ENTER_CS;

	/* ML[ApPbgo܂B */
	retval = remove_queue(mac.rxque, packet, maxlength);

LEAVE_CS;

	/* Ȃ΁ApPbgTCY܂B */
	if(retval >= 0) {

		/* ŏt[̃pPbǵAML[֊i[Ɏ̂ĂĂ͂B */
		if(retval < 60) {
			DIE();
		}

		/* őt[߂̃pPbǵAML[֊i[Ɏ̂ĂĂ͂B */
		if(retval > 1514) {
			DIE();
		}
	}

	return retval;
}

int
mac_send(const void* packet/*[length]*/, int length)
{
	int retval;

	if(length < 0) {
		DIE();
	}

	/* ŏt[TCYւ̊ǵAML[oɍs܂B */
	if(length < 60) {
		/** no job **/
	}

	/* 2005/08/17݂̎ł́AőTCY߂̗v̓G[~Ƃ܂B */
	if(length > 1514) {
		DIE();
	}

ENTER_CS;

	/* pPbgAML[֒ǉ܂B */
	retval = append_queue(mac.txque, packet, length);
	if(retval >= 0) { /* ǉ */

		/* M~Ȃ... */
		if(!mac.txing) {

			/* oJn܂Bȍ~́Aep2_txdone()ŘA܂B */
			mac.txing = ep2_txsub();
			ASSERT(mac.txing);
		}

	} else { /* ǉs */

		mac.txovr++; /* MI[o[ */
	}

LEAVE_CS;

	return retval;
}

void
mac_address(void* address/*[6]*/)
{
ENTER_CS;

	memcpy(address, mac.address, 6);

LEAVE_CS;
}

void
mac_stat(MACSTAT* stat)
{
ENTER_CS;

	stat->rxok  = mac.rxok;
	stat->rxerr = mac.rxerr;
	stat->rxovr = mac.rxovr;
	stat->txok  = mac.txok;
	stat->txerr = mac.txerr;
	stat->txovr = mac.txovr;

LEAVE_CS;
}
